from ..helpers.point import Point, CenterPosition


class Roi:
    """ A class to manage ROIs in DICOM images

    Args:
        center : The coordinates for the center position (x-, y-, and z-index) of the ROI

    Attributes:
        Center : The coordinates for the center position (x-, y-, and z-index) of the ROI

    """
    def __init__(self, center: CenterPosition):
        if not isinstance(center, Point) and not isinstance(center, dict):
            raise TypeError("The center must be given as a point or a dict")

        if isinstance(center, Point):
            self.Center = center
        else:
            if 'x' not in center.keys() or 'y' not in center.keys() or 'z' not in center.keys():
                raise ValueError(("The center dict must be on the form dict(x: Union[float, int], y: Union[float, int],"
                                  " z: Optional[Union[float, int]}"))
            z = center.get('z')
            if z is not None:
                z = float(z)
            self.Center = Point(x=float(center.get('x')), y=float(center.get('y')), z=z)

        if sum([self.Center.x is None, self.Center.y is None]) > 0:
            raise ValueError("At least the x and y positions of the center must be specified")



import numpy as np
from pathlib import Path
import pydicom
from pydicom import FileDataset
from typing import List, Optional

from .dicom_series import DicomSeries
from ..helpers.pixel_data import get_pixel_array
from ..helpers.voxel_data import VoxelData


class ProjectionSeries(DicomSeries):
    """ A class to manage 2D DICOM images, e.g., from conventional X-rays, mammography, panoramic, intraoral etc.

    This class only handles one image per series, as is usually the case for this kind of images.

    Args:
        file: Path object for the file that is to be imported
        dcm: A pydicom FileDataset object containing the file

    Attributes:
        kV: Tube voltage used in the image acquisition in kV
        mA: Tube current used in the image acquisition in mA
        Modality: The modality of the image, e.g., IO, PX, MG, DX, CR, etc.
        Manufacturer: The name of the manufacturer as specified in the DICOM file
        ManufacturersModelName: The model name specified by the manufacturer as given in the DICOM file

    """
    def __init__(self, file: Path, dcm: Optional[FileDataset] = None):
        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if 'SeriesInstanceUID' not in dcm:
            raise ValueError("The DICOM file does not contain a series instance UID")

        super().__init__(series_instance_uid=dcm.SeriesInstanceUID)

        self.Modality = dcm.Modality
        self.Manufacturer: Optional[str] = None
        if 'Manufacturer' in dcm:
            self.Manufacturer = dcm.Manufacturer

        self.ManufacturersModelName: Optional[str] = None
        if 'ManufacturerModelName' in dcm:
            self.ManufacturersModelName = dcm.ManufacturerModelName
        if self.Modality == 'IO' and 'DetectorManufacturerModelName' in dcm:
            self.ManufacturersModelName = dcm.DetectorManufacturerModelName

        self.kV: Optional[List[float]] = []
        self.mA: Optional[List[float]] = []
        self.ImageVolume: Optional[List[np.ndarray]] = []

        self.add_file(file=file, dcm=dcm)

    def add_file(self, file: Path, dcm: Optional[FileDataset] = None):
        """ Add a file to the objects list of files

        First performs a check that the file path is of a path object and that it has the same series instance UID as
        the class object

        Args:
            file: Path to where the file to be added is stored on disc
            dcm: The DICOM-file imported to a FileDataset object

        Raises:
            InvalidDicomError: If the given file is not a valid DICOM file
            ValueError: If the file does not have the same study instance UID as the StudyInstanceUID of the object

        """
        if not isinstance(file, Path):
            raise TypeError("file is not a Path-object")

        super().add_file(file=file, dcm=dcm)

        if len(self.CompleteMetadata) == len(self.FilePaths):
            # Skip out because the file has already been added
            return

        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if 'PixelSpacing' in dcm:
            self.VoxelData.append(VoxelData(x=float(dcm.PixelSpacing[1]),
                                            y=float(dcm.PixelSpacing[0]),
                                            z=None))
        else:
            # Assume pixel size is set in Detector Element Spacing tag (0018, 7022)
            self.VoxelData.append(VoxelData(x=float(dcm.DetectorElementSpacing[1]),
                                            y=float(dcm.DetectorElementSpacing[0]),
                                            z=None))
        if 'KVP' in dcm:
            self.kV.append(float(dcm.KVP))

        if 'XRayTubeCurrent' in dcm:
            self.mA.append(float(dcm.XRayTubeCurrent))
        elif 'XRayTubeCurrentInmA' in dcm:
            self.mA.append(float(dcm.XRayTubeCurrentInmA))
        elif 'XRayTubeCurrentInuA' in dcm:
            self.mA.append(float(dcm.XRayTubeCurrentInuA) / 1000)

        # Remove pixel data part of dcm to decrease memory used for the object
        if 'PixelData' in dcm:
            try:
                del dcm[0x7FE00010]
            except Exception as e:
                pass

        self.CompleteMetadata.append(dcm)

    def import_image(self) -> None:
        """ Import the pixel data into the ImageVolume property

        Returns:

        """
        self.ImageVolume = []
        for fp in self.FilePaths:
            dcm = pydicom.dcmread(str(fp.absolute()))
            self.ImageVolume.append(get_pixel_array(dcm=dcm))



import logging
import numpy as np
from pydicom import Dataset


log = logging.getLogger(__name__)


def get_pixel_array(dcm: Dataset) -> np.ndarray:
    """ Take a DICOM dataset, extract the pixels, rescale the values according to metadata. Return the extracted image
    as a numpy ndarray of int16 values

    Args:
        dcm: The DICOM dataset from which the image should be extracted

    Returns:
        Extracted image as a numpy ndarray of int16 numbers

    """
    px = dcm.pixel_array.astype(np.int16)
    if 'RescaleSlope' in dcm:
        log.debug('Rescaling slope of pixel array')
        px *= np.int16(dcm.RescaleSlope)
    if 'RescaleIntercept' in dcm:
        log.debug('Rescaling intercept of pixel array')
        px += np.int16(dcm.RescaleIntercept)

    return px



from pathlib import Path
import pydicom
from pydicom.errors import InvalidDicomError
from typing import Dict, Optional

from .dicom_study import DicomStudy
from .ct import CtSeries


def import_dicom_from_folder(folder: Path, recursively: bool = True) -> Dict[str, DicomStudy]:
    """ Go through a folder and import all valid DICOM images found

    Args:
        folder: Path object of the folder to search for DICOM files
        recursively: Specification if the folder should be search recursively. Defaults to True

    Raises:
        TypeError: If the given folder is not a Path object
        ValueError: If the given folder is not a directory
        ValueError: If no valid DICOM files were found in the search of the given folder

    Returns:
        A dictionary on the form {<study-instance-uid>: <DicomStudy object>}

    """
    if not isinstance(folder, Path):
        raise TypeError("folder must be a Path object")
    if not folder.is_dir():
        raise ValueError("The given folder is not a directory")

    files = folder.iterdir()
    if recursively:
        files = folder.rglob("*")

    dicom_study_list = dict()

    for fp in files:
        if not fp.is_file():
            continue

        try:
            dcm = pydicom.dcmread(fp=str(fp.absolute()), stop_before_pixels=True)
        except InvalidDicomError as e:
            continue

        if dcm.StudyInstanceUID not in dicom_study_list:
            dicom_study_list[dcm.StudyInstanceUID] = DicomStudy(study_instance_uid=dcm.StudyInstanceUID)

        dicom_study_list[dcm.StudyInstanceUID].add_file(fp, dcm=dcm)

    if not len(dicom_study_list):
        raise ValueError("The given folder contains no valid DICOM files")

    return dicom_study_list


def import_dicom_file(file: Path) -> DicomStudy:
    """ Import a DICOM file into a DicomStudy object

    Args:
        file: Path to the file to import

    Raises:
        TypeError: If the given file is not a Path object
        ValueError: If the given file path is not a valid file
        InvalidDicomError: If the given file is not a valid DICOM file

    Returns:
        DicomStudy object with the file added to it

    """
    if not isinstance(file, Path):
        raise TypeError("file must be a Path object")
    if not file.is_file():
        raise ValueError("File is not a valid file")

    try:
        dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)
    except InvalidDicomError:
        raise

    output = DicomStudy(study_instance_uid=dcm.StudyInstanceUID)
    output.add_file(file=file, dcm=dcm)

    if isinstance(output.Series[0], CtSeries):
        output.Series[0].import_image_volume()
    else:
        output.Series[0].import_image()

    return output



import numpy as np
from pathlib import Path
import pydicom
from pydicom import FileDataset
from typing import List, Optional
from ..helpers.voxel_data import VoxelData


class DicomSeries:
    """ A class to manage DICOM files connected by a Series Instance UID

    Args:
        series_instance_uid: Series instance UID of the object to be created

    Attributes:
        SeriesInstanceUid: Series instance UID of the object
        FilePaths: Paths to the files added to the object
        CompleteMetadata: The complete set of metadata for the added files
        VoxelData: Voxel size information for included image files
        ImageVolume: The Image volume of the DICOM series
        Mask: A mask of the same dimension as the image volume to apply to the image volume

    """
    def __init__(self, series_instance_uid: str):
        if not isinstance(series_instance_uid, str):
            raise TypeError("series_instance_uid must be a string")
        self.FilePaths: List[Path] = []

        # Metadata
        self.SeriesInstanceUid: str = series_instance_uid
        self.SeriesDescription: Optional[str] = None
        self.CompleteMetadata: List[Optional[FileDataset]] = []
        self.VoxelData: List[VoxelData] = []

        self.ImageVolume: Optional[np.ndarray] = None
        self.Mask: Optional[np.ndarray] = None

    def add_file(self, file: Path, dcm: Optional[FileDataset] = None):
        """ Add a file to the objects list of files

        First performs a check that the file is a valid DICOM file and that it belongs to the object/series

        Args:
            file: Path to where the file to be added is stored on disk
            dcm: The DICOM-file imported to a FileDataset object

        Raises:
            ValueError: if SeriesInstanceUID of the file is not the same as the SeriesInstanceUid attribute

        """
        if any([True if obj == file else False for obj in self.FilePaths]):
            # Return None since the file is already in hte volume
            return

        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if dcm.SeriesInstanceUID != self.SeriesInstanceUid:
            msg = f"Wrong SeriesInstanceUID. Expected: {self.SeriesInstanceUid}; Input: {dcm.SeriesInstanceUID}"
            raise ValueError(msg)

        if 'SeriesDescription' in dcm:
            self.SeriesDescription = dcm.SeriesDescription

        self.FilePaths.append(file)




from numpy import (abs, absolute, add, arctan, array, ceil, clip, divide, exp, median, multiply, ones, pi, power, sort,
                   sqrt, subtract, sum, vectorize, zeros)
from scipy import linalg
from scipy.optimize import curve_fit
from typing import List, Optional, Tuple, Union

from .point import Point, IntPoint

LINE_COORDINATE = Union[int, float]
FIT_DATA = Union


class Line:
    def __init__(self, x0: LINE_COORDINATE, y0: LINE_COORDINATE, x1: LINE_COORDINATE, y1: LINE_COORDINATE):
        if not all([(isinstance(obj, int) or isinstance(obj, float)) for obj in [x0, y0, x1, y1]]):
            raise TypeError("All line coordinates must be either integers or floats")

        self.slope = (y1 - y0) / (x1 - x0)
        self.constant = y0 - self.slope * x0
        self.angle = arctan(self.slope) * 180 / pi


def get_shortest_line_point_dist(line: Line, point: Union[Point, IntPoint]):
    if not isinstance(line, Line):
        raise TypeError("line must be a Line object")

    if not isinstance(point, Point) and not isinstance(point, IntPoint):
        raise TypeError("point must be either a Point or IntPoint instance")

    distance = absolute(line.slope * point.x + line.constant - point.y) / sqrt(line.slope ** 2 + (-1) ** 2)
    if line.angle <= 45 or line.angle >= 135:
        if point.y < (line.slope * point.x + line.constant):
            distance = -distance
    else:
        if point.x < (point.y - line.constant) / line.slope:
            distance = -distance

    return distance


def third_deg_polynomial_fit_func(x, a, b, c, d):
    return add(multiply(a, power(x, 3)), add(multiply(b, power(x, 2)), add(multiply(c, x), d)))


def get_third_degree_polynomial_fit(x: List[Union[float, int]], y: List[Union[float, int]]):
    """ Fits data to a 3rd degree polynomial on the format y = a*x^3 + b*x^2 + c*x + d. Returns y-values calculated with
    the optimized function and the z, b, c, and d parameters in a dict.

    Args:
        x: Values on the x-axis
        y: Noisy y-values to fit to the function

    Returns:
        A dictionary on with the keys a, b, c, d and y, where y is the optimized y values.

    """
    popt, pcov = curve_fit(f=third_deg_polynomial_fit_func, xdata=x, ydata=y)
    optimized_y = third_deg_polynomial_fit_func(x, *popt)

    return dict(a=popt[0], b=popt[1], c=popt[2], d=popt[3], y=optimized_y)


def _bell_kernel_function(x: Union[int, float], x0: Union[int, float], tau: Optional[float] = 0.005):
    return exp(-divide(power(subtract(x, x0), 2), multiply(2, tau)))


def _test_function(x, x1, ind):
    output = sort(abs(subtract(x, x1)))[ind]
    print(output)
    return output


def lowess(x: List[Union[float, int]], y: List[Union[float, int]], f: Optional[float] = 2. / 3.,
           iter: Optional[int] = 3):
    """lowess(x, y, f=2./3., iter=3) -> yest
    Lowess smoother: Robust locally weighted regression.
    The lowess function fits a nonparametric regression curve to a scatterplot.
    The arrays x and y contain an equal number of elements; each pair
    (x[i], y[i]) defines a data point in the scatterplot. The function returns
    the estimated (smooth) values of y.
    The smoothing span is given by f. A larger value for f will result in a
    smoother curve. The number of robustifying iterations is given by iter. The
    function will run faster with a smaller number of iterations.
    """
    n = len(x)
    r = int(ceil(f * n))
    tf = vectorize
    h = [sort(abs(subtract(x, x[i])))[r] for i in range(n)]
    w = clip(abs(divide((subtract(x[:, None], x[None, :])), h)), 0.0, 1.0)
    w = power(power(subtract(1, w), 3), 3)
    yest = zeros(n)
    delta = ones(n)
    for iteration in range(iter):
        for i in range(n):
            weights = multiply(delta, w[:, i])
            b = array([sum(multiply(weights, y)), sum(multiply(weights, multiply(y, x)))])
            A = array([[sum(weights), sum(multiply(weights, x))],
                       [sum(weights * x), sum(multiply(multiply(weights, x), x))]])
            beta = linalg.solve(A, b)
            yest[i] = add(beta[0], multiply(beta[1], x[i]))

        residuals = subtract(y, yest)
        s = median(abs(residuals))
        delta = clip(residuals / multiply(6.0, s), -1, 1)
        delta = power(subtract(1, power(delta, 2)), 2)

    return yest



import logging
import numpy as np
from pathlib import Path
import pydicom
from pydicom import FileDataset
from scipy import ndimage
from scipy.ndimage import center_of_mass
from skimage import morphology
from typing import List, Optional

from .dicom_series import DicomSeries
from ..helpers.pixel_data import get_pixel_array
from ..helpers.voxel_data import VoxelData
from ..helpers.patient_centering import PatientMassCenter, PatientGeometricalOffset


log = logging.getLogger(__name__)


class CtSeries(DicomSeries):
    """ A class to manage DICOM files from CT connected by a Series Instance UID

    Args:
        series_instance_uid: Series instance UID of the object to be created

    Attributes:
        kV: Tube voltage used for each image in kV
        mA: Tube current used per image in mA
        SlicePosition: List of slice locations (in mm)
        Manufacturer: The name of the manufacturer of the CT machine
        ManufacturersModelName: The model name specified by the manufacturer as given in the DICOM file
        Mask: A mask covering the patient, created in the get_patient_mask method
        MaskSuccess: A boolean saying if the patient mask creationg was successful or not
        PatientClipped: A boolean specifying if the masked patient contour touches the any of the edges of the image
                        volume
        PatientMassCenterImage: A list of the masked patient mass center for each image
        PatientMassCenterVolume: The masked patient volume mass center
        PatientGeometricalOffset: The patient volume geometrical offset from the scanner isocenter
        MeanHuPatientImage: Mean HU value of the masked patient for each image
        MedianHuPatientImage: Median HU value of the masked patient for each image
        MeanHuPatientVolume: Mean HU value of the masked patient volume
        MedianHuPatientVolume: Median HU value of the masked patient volume

    """
    def __init__(self, series_instance_uid: str):
        super().__init__(series_instance_uid=series_instance_uid)
        self.kV: Optional[List[float]] = None
        self.mA: Optional[List[Optional[float]]] = None
        self.SlicePosition: Optional[List[float]] = []
        self.Manufacturer: Optional[str] = None
        self.ManufacturersModelName: Optional[str] = None

        # Image volume data
        self.Mask: Optional[np.ndarray] = None
        self.MaskSuccess: Optional[bool] = None
        self.PatientClipped: Optional[bool] = None

        # Calculated image information
        self.PatientMassCenterImage: Optional[List[PatientMassCenter]] = None
        self.PatientMassCenterVolume: Optional[PatientMassCenter] = None
        self.PatientGeometricalOffset: Optional[PatientGeometricalOffset] = None
        self.MeanHuPatientImage: Optional[List[float]] = None
        self.MedianHuPatientImage: Optional[List[float]] = None
        self.MeanHuPatientVolume: Optional[float] = None
        self.MedianHuPatientVolume: Optional[float] = None

    def add_file(self, file: Path, dcm: Optional[FileDataset] = None) -> None:
        """ Add a file to the objects list of files.

        First performs a check that the file is a valid DICOM file and that it is a CT file.
        List properties are emptied to prevent mismatch between them, the FilePaths and SlicePosition

        Raises:
            ValueError: If supplied file is not a CT image

        """
        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if dcm.Modality.upper() != "CT":
            raise ValueError(f"The supplied file is not a CT image. (supplied modality: {dcm.Modality}")

        self.ImageVolume = None
        self.kV = []
        self.mA = []
        self.PatientMassCenterImage = []
        self.PatientMassCenterVolume = None
        self.PatientGeometricalOffset = None
        self.MeanHuPatientVolume = None
        self.MeanHuPatientImage = []
        self.MedianHuPatientVolume = None
        self.MedianHuPatientImage = []
        self.Mask = None
        self.MaskSuccess = None
        self.PatientClipped = None

        super().add_file(file=file, dcm=dcm)
        self.Manufacturer = dcm.Manufacturer
        self.ManufacturersModelName = dcm.ManufacturerModelName
        self.SlicePosition.append(float(dcm.SliceLocation))

        # Reorder lists according to slice positions
        file_order = [ind for ind in list(np.argsort(np.array(self.SlicePosition)))]
        self.FilePaths = [self.FilePaths[ind] for ind in file_order]
        self.SlicePosition = [self.SlicePosition[ind] for ind in file_order]

    def import_image_volume(self) -> None:
        """ Import the files in the CtVolume and insert them into the ImageVolumeProperty. Also add metadata for each
        image into the respective property.
        """
        # Remove any previously imported image volume
        self.ImageVolume = None
        self.kV = []
        self.mA = []
        self.SlicePosition = []
        self.PatientMassCenterImage = []

        for ind, fp in enumerate(self.FilePaths):
            dcm = pydicom.dcmread(fp=str(fp.absolute()))
            px = get_pixel_array(dcm=dcm)

            if self.ImageVolume is None:
                self.ImageVolume = np.empty((px.shape[0], px.shape[1], len(self.FilePaths)))

            self.ImageVolume[:, :, ind] = px
            self.kV.append(float(dcm.KVP))

            if 'XRayTubeCurrent' in dcm:
                self.mA.append(float(dcm.XRayTubeCurrent))
            elif 'XRayTubeCurrentInmA' in dcm:
                self.mA.append(float(dcm.XRayTubeCurrentInmA))
            elif 'XRayTubeCurrentInuA' in dcm:
                self.mA.append(float(dcm.XRayTubeCurrentInuA) / 1000)
            else:
                self.mA.append(None)

            self.SlicePosition.append(float(dcm.SliceLocation))

            self.VoxelData.append(VoxelData(x=float(dcm.PixelSpacing[1]), y=float(dcm.PixelSpacing[0]),
                                            z=float(dcm.SliceThickness)))

            # Remove pixel data part of dcm to decrease memory used for the object
            try:
                del dcm[0x7FE00010]
            except Exception as e:
                pass

            self.CompleteMetadata.append(dcm)

    def get_patient_mask(self, threshold: Optional[int] = -500, remove_table: Optional[bool] = False):
        """ Segment the ImageVolume to find the patient/phantom in the images.

        Args:
            threshold: HU value to use as threshold. Defaults to -500
            remove_table: Specify if the CT table should be removed from the image. Defaults to False

        Raises:
            TypeError: If threshold is not an integer
            ValueError: If there is not image volume to segment

        """
        if not isinstance(threshold, int):
            raise TypeError("The threshold must be given as an integer")

        if self.ImageVolume is None:
            self.import_image_volume()
            if self.ImageVolume is None:
                raise ValueError("Found no image volume to segment")

        self.Mask = np.zeros(self.ImageVolume.shape)
        self.Mask[self.ImageVolume >= threshold] = 1

        if remove_table:
            # Remove the table by eroding and dilating the image volume
            if self.Mask.shape[2] > 2:
                self.Mask = morphology.binary_erosion(image=self.Mask, selem=morphology.cube(width=3))
            else:
                for i in range(self.Mask.shape[2]):
                    self.Mask[:, :, i] = morphology.binary_erosion(image=self.Mask[:, :, i],
                                                                   selem=morphology.disk(radius=3))

            self.Mask, nb_labels = ndimage.label(self.Mask)

            central_position = [int(np.floor(np.divide(float(self.Mask.shape[0]), 2.0))),
                                int(np.floor(np.divide(float(self.Mask.shape[1]), 2.0))),
                                int(np.floor(np.divide(float(self.Mask.shape[2]), 2.0)))]

            central_blob = np.max(self.Mask[(central_position[0] - 2):(central_position[0] + 3),
                                  (central_position[1] - 2):(central_position[1] + 3),
                                  central_position[2]])

            self.Mask[self.Mask != central_blob] = 0
            self.Mask[self.Mask == central_blob] = 1

            if self.Mask.shape[2] > 2:
                self.Mask = morphology.binary_dilation(image=self.Mask, selem=morphology.cube(width=3))
            else:
                for i in range(self.Mask.shape[2]):
                    self.Mask[:, :, i] = morphology.binary_dilation(image=self.Mask[:, :, i],
                                                                    selem=morphology.disk(radius=3))

        for i in range(self.Mask.shape[2]):
            self.Mask[:, :, i] = ndimage.binary_fill_holes(self.Mask[:, :, i]).astype(int)

        if np.sum(self.Mask) > 0:
            log.info("Axial images segmented successfully")
            self.MaskSuccess = True

        self.Mask = self.Mask.astype(bool)
        tmp_mass_center = [center_of_mass(input=self.Mask[:, :, ind]) for ind in range(self.Mask.shape[2])]
        self.PatientMassCenterImage = [PatientMassCenter(x=obj[1], y=obj[0]) for obj in tmp_mass_center]

        if self.Mask.shape[2] > 1:
            tmp_mass_center = center_of_mass(input=self.Mask)
            self.PatientMassCenterVolume = PatientMassCenter(x=tmp_mass_center[1], y=tmp_mass_center[0],
                                                             z=tmp_mass_center[2])
        else:
            self.PatientMassCenterVolume = self.PatientMassCenterImage[0]

        # Try to get the patient geometrical offset
        try:
            self._get_patient_geometrical_offset()
        except Exception as e:
            log.warning("Could not calculate patient geometrical offset", e)

        # Check if patient is clipped
        self.PatientClipped = any([
            np.sum(self.Mask[0, :, :]) > 0,
            np.sum(self.Mask[-1, :, :]) > 0,
            np.sum(self.Mask[:, 0, :]) > 0,
            np.sum(self.Mask[:, -1, :]) > 0,
        ])

        # Calculate mean and median values
        tmp_masked_image = np.ma.array(self.ImageVolume, mask=np.logical_not(self.Mask))

        self.MeanHuPatientImage = list(tmp_masked_image.mean(axis=(0, 1)))
        self.MedianHuPatientImage = list(np.ma.median(tmp_masked_image, axis=(0, 1)))
        self.MeanHuPatientVolume = tmp_masked_image.mean(axis=None)
        self.MedianHuPatientVolume = np.ma.median(tmp_masked_image, axis=None)

    def _get_patient_geometrical_offset(self):
        """ Calculate the patient/phantom geometrical offset from isocenter

        Raises:
            ValueError: If the DICOM header does not contain all tags required for the calculation

        """
        log.info("Calculating phantom geometrical offset")

        for ind, mass_center in enumerate(self.PatientMassCenterImage):
            if 'ImagePositionPatient' in self.CompleteMetadata[ind]:
                geometrical_center = (round(mass_center.y, 1), round(mass_center.x, 1))
                image_position = (float(self.CompleteMetadata[ind].ImagePositionPatient[0]),
                                  float(self.CompleteMetadata[ind].ImagePositionPatient[1]))

                if self.Manufacturer.upper() in ['SIEMENS', 'PHILIPS']:
                    # Recalculate image position due to incorrect specifications
                    image_position[1] += float(self.CompleteMetadata[ind].TableHeight)

                if self.PatientGeometricalOffset is None:
                    self.PatientGeometricalOffset = []

                self.PatientGeometricalOffset.append(PatientGeometricalOffset(
                    x=float(image_position[0]) + geometrical_center[1] + self.VoxelData[ind].x,
                    y=float(image_position[1]) + geometrical_center[0] + self.VoxelData[ind].y,
                ))
                continue

            if 'DataCollectionCenterPatient' in self.CompleteMetadata[ind] and\
                    'ReconstructionTargetCenterPatient' in self.CompleteMetadata[ind]:
                log.debug(("Calculating diff from table based on DataCollectionCenterPatient and "
                           "ReconstructionTargetCenterPatient"))
                diff_table_x_mm = float(self.CompleteMetadata[ind].ReconstructionTargetCenterPatient[0]) - float(
                    self.CompleteMetadata[ind].DataCollectionCenterPatient[0])
                diff_table_y_mm = float(self.CompleteMetadata[ind].ReconstructionTargetCenterPatient[1]) - float(
                    self.CompleteMetadata[ind].DataCollectionCenterPatient[1])

            elif self.Manufacturer.upper() in ['GE MEDICAL SYSTEMS'] and self.CompleteMetadata[ind][0x431031]:
                log.debug(f"Calculating diff from table for machine from {self.Manufacturer}")
                diff_table_x_mm = float(self.CompleteMetadata[ind][0x431031].value[0])
                diff_table_y_mm = float(self.CompleteMetadata[ind][0x431031].value[1])

            elif self.Manufacturer.upper() in ['TOSHIBA'] and self.CompleteMetadata[ind][0x70051007]:
                log.debug(f"Calculating diff from table for machine from {self.Manufacturer}")
                tmp = [float(i) for i in self.CompleteMetadata[ind][0x70051007].value.decode('utf-8').split('\\')]
                diff_table_x_mm = (tmp[0] - self.CompleteMetadata[ind].Columns / 2.0) * self.VoxelData[ind].x
                diff_table_y_mm = (tmp[1] - self.CompleteMetadata[ind].Rows / 2.0) * self.VoxelData[ind].y

            else:
                raise ValueError("The DICOM header does not contain all required tags")

            # Calculate patient offset
            image_center_x = self.CompleteMetadata[ind].Columns / 2.0
            image_center_y = self.CompleteMetadata[ind].Rows / 2.0

            log.debug(f"Image center: x={image_center_x}, y={image_center_y}")

            geometrical_center = (round(mass_center[0], 1), round(mass_center[1], 1))

            diff_geom_x_center_mm = (geometrical_center[1] - image_center_x) * self.VoxelData[ind].x
            diff_geom_y_center_mm = (geometrical_center[0] - image_center_y) * self.VoxelData[ind].y

            self.PatientGeometricalOffset.append(PatientGeometricalOffset(
                x=diff_table_x_mm - diff_geom_x_center_mm,
                y=diff_table_y_mm - diff_geom_y_center_mm
            ))



import numpy as np
from scipy.stats import sem
from skimage import color

from .roi import Roi, CenterPosition
from ..helpers.point import IntPoint
from ..helpers.voxel_data import VoxelData


VALID_COLOURS = {
    'SkyBlue': (135, 206, 235),
    'MediumBlue': (0, 0, 205)
}


class SquareRoi(Roi):
    """
    A class for square ROIs in images.

    Attributes
    ----------
    center : CenterPosition
        The coordinates for the center position (x-, y-, and z-index) of the ROI
    height : float
        ROI height in mm
    width : float
        ROI width in mm
    pixel_size : VoxelData
        Pixel size/spacing in mm
    resize_too_big_roi : bool
        Resize a ROI that reaches outside of the image

    Methods
    -------
    get_mean(image)
        Returns the mean of the pixel values in the ROI
    get_stdev(image)
        Returns the standard deviation of the pixel values in the ROI
    get_sum(image, axis=1)
        Returns the sum of the pixel values in the ROI
    get_std_error_of_the_mean(image)
        Returns the standard error of the mean of the pixel values in the ROI
    add_roi_to_image(image, roi_color='SkyBlue')
        Returns the image with the ROI added, coloured in the requested colour
    get_roi_part_of_image(image)
        Returns a numpy ndarray only containing the part of the image that is contained in the ROI
    """

    def __init__(self, center: CenterPosition, height: float, width: float, pixel_size: VoxelData,
                 resize_too_big_roi: bool = False):
        super().__init__(center=center)
        self.Height: float = height
        self.Width: float = width
        self.ResizeTooBigRoi: bool = resize_too_big_roi

        height_pixels = int(round(self.Height / pixel_size.y))
        width_pixels = int(round(self.Width / pixel_size.x))

        if height_pixels < 1 or width_pixels < 1:
            raise ValueError((f'Too small ROI size specified. Both width ({width_pixels}) and height ({height_pixels}) '
                              f'must be at least 1 pixel'))

        height_pixels = int(round((self.Height - pixel_size.y) / pixel_size.y))
        if height_pixels < 0:
            height_pixels = 0
        width_pixels = int(round((self.Width - pixel_size.x) / pixel_size.x))
        if width_pixels < 0:
            width_pixels = 0

        self.UpperLeft: IntPoint = IntPoint(
            x=int(np.floor(self.Center.x - (width_pixels / 2) + 0.5)),
            y=int(np.floor(self.Center.y - (height_pixels / 2) + 0.5))
        )
        # Validate corner does not have negative index
        if self.UpperLeft.x < 0 or self.UpperLeft.y < 0:
            if not resize_too_big_roi:
                raise ValueError(("The upper left corner of the ROI reaches outside of the image "
                                  f"(x = {self.UpperLeft.x}, y = {self.UpperLeft.y})"))
            if self.UpperLeft.x < 0:
                self.UpperLeft.x = 0
            if self.UpperLeft.y < 0:
                self.UpperLeft.y = 0

        self.LowerLeft: IntPoint = IntPoint(
            x=int(np.floor(self.Center.x - (width_pixels / 2) + 0.5)),
            y=int(np.floor(self.Center.y + (height_pixels / 2) + 0.5))
        )

        self.UpperRight: IntPoint = IntPoint(
            x=int(np.floor(self.Center.x + (width_pixels / 2) + 0.5)),
            y=int(np.floor(self.Center.y - (height_pixels / 2) + 0.5)))

        self.LowerRight: IntPoint = IntPoint(
            x=int(np.floor(self.Center.x + (width_pixels / 2) + 0.5)),
            y=int(np.floor(self.Center.y + (height_pixels / 2) + 0.5)))

    def _check_roi_placement(self, image: np.ndarray) -> (int, int):
        """ Checks that the ROI is inside the given image. If not raise Value error or if ResizeTooBigRoi == True,
        calculate new ending indexes.

        :param image: Numpy ndarray containing the image
        :return: The resulting ending indexes as a x, y
        """
        image_shape = image.shape
        if self.UpperLeft.x > image_shape[1] or self.UpperLeft.y > image_shape[0]:
            raise ValueError("Entire image outside of the image")

        x2 = self.LowerRight.x
        if self.LowerRight.x >= image_shape[1]:
            if not self.ResizeTooBigRoi:
                raise ValueError("The ROI reaches outside of the image")
            x2 = image_shape[1]

        y2 = self.LowerRight.y
        if self.LowerRight.y >= image_shape[0]:
            if not self.ResizeTooBigRoi:
                raise ValueError("The ROI reaches outside of the image")
            y2 = image_shape[0]

        return x2, y2

    def get_mean(self, image: np.ndarray) -> np.ndarray:
        """ Calculates the mean of the pixel values contained in the ROI from the input image

        :param image: Numpy ndarray containing the image
        :return: The mean of the pixel values from the part of the input image contained in the ROI
        """
        x2, y2 = self._check_roi_placement(image=image)

        return np.mean(image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1])

    def get_stdev(self, image: np.ndarray) -> np.ndarray:
        """ Calculates the standard deviation of the pixel values contained in the ROI from the input image

        :param image: Numpy ndarray containing the image
        :return: The standard deviation of the pixel values from the part of the input image contained in the ROI
        """
        x2, y2 = self._check_roi_placement(image=image)

        return np.std(image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1])

    def get_sum(self, image: np.ndarray, axis: int = 1):
        """ Calculates the sum of the pixel values contained in the ROI from the input image

        :param image: Numpy ndarray containing the image
        :param axis: The axis over which to calculate the sum
        :return: The sum of the square ROI applied to the supplied image over the specified axis
        """
        x2, y2 = self._check_roi_placement(image=image)

        return np.sum(a=image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1], axis=axis)

    def get_std_error_of_the_mean(self, image: np.ndarray) -> np.ndarray:
        """ Calculates the standard error of the mean of the pixel values contained in the ROI from the input image

        :param image: Numpy ndarray containing the image
        :return: The standard error of the mean of the pixel values from the part of the input image contained in the
        ROI
        """
        x2, y2 = self._check_roi_placement(image=image)

        return sem(image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1].flatten())

    def add_roi_to_image(self, image: np.ndarray, roi_color: str = 'SkyBlue') -> np.ndarray:
        """ Adds the ROI to the given image using the color specified.

        :param image: Numpy ndarray containing the image
        :param roi_color: The color to use for the ROI
        :return: The input image with the ROI added
        """
        roi_color = VALID_COLOURS.get(roi_color)

        if roi_color is None:
            raise ValueError(f'The colour must be one of {",".join(VALID_COLOURS.keys())}')

        x2, y2 = self._check_roi_placement(image=image)

        image = color.gray2rgb(image)

        image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1] = roi_color
        return image

    def get_roi_part_of_image(self, image: np.ndarray) -> np.ndarray:
        """ Extracts the part of the image contained in the ROI

        :param image: Numpy ndarray containing the image
        :return: The part of the input image contained in the square ROI as a numpy ndarray
        """
        x2, y2 = self._check_roi_placement(image=image)

        return image[self.UpperLeft.y:y2 + 1, self.UpperLeft.x:x2 + 1]



from dataclasses import dataclass
import logging
from typing import List

import numpy as np
from scipy.signal import savgol_filter
from scipy.stats import norm
from skimage.feature import canny
from skimage.transform import probabilistic_hough_line
from skimage.filters import gaussian

from ..helpers.geometry import Line, get_shortest_line_point_dist
from ..helpers.point import IntPoint
from ..roi.square_roi import SquareRoi

logger = logging.getLogger(__name__)


@dataclass
class MtfCurve:
    """Class for holding an MTF-curve"""
    frequency: List[int]
    mtf: np.ndarray


def mtf_bar_pattern(max_hu: float, bg: float, sd: List[float], sd_bg: float) -> List[float]:
    """ Implement the method by Droege and Morin (1982) Medical Physics. Determine MTF from bar pattern data

    See wiki for a detailed explanation of the method.

    Args:
        max_hu: The maximum CT number in the bar pattern ROI
        bg: The background CT number
        sd: A list of standard deviations over the line pair patterns for the different frequencies
        sd_bg: The standard deviation of a homogeneous background ROI

    Returns:
        A list of the MTF value in the same order as the supplied standard deviations

    """
    if not isinstance(max_hu, float):
        raise TypeError("max_hu must be a float")

    if not isinstance(bg, float):
        raise TypeError("bg must be a float")

    if not isinstance(sd, List) or not any([isinstance(el, float) for el in sd]):
        raise TypeError("sd must be a list of floats")

    if not isinstance(sd_bg, float):
        raise TypeError("sd_bg must be a float")

    m0 = np.divide(np.abs(np.subtract(max_hu, bg)), 2)
    n = sd_bg
    m = [np.sqrt(i ** 2 - n ** 2) if i > n else 0 for i in sd]

    mtf = [np.multiply(np.divide(np.multiply(np.pi, np.sqrt(2.0)), 4.0), np.divide(msa, m0)) for msa in m]

    return mtf


def mtf_edge_phantom(image: np.ndarray, roi: SquareRoi) -> MtfCurve:
    """ Calculate the MTF from an image of an edge phantom. Find the edge using a canny filter in combination with the
    Hough transform. The raw ESF is smoothed using a Savitzky-Golay filter with a 4th order polynomial.

    Args:
        image: A numpy ndarray of the image containing the edge phantom
        roi: A square ROI object specifying the region in which the edge phantom is contained

    Returns:
        An :obj:`MtfCurve <dicom_image_tools.image_quality_mtf.MtfCurve>` object instance with the frequency in
        cycles/pixel

    """
    # Find edge
    edge_image = roi.get_roi_part_of_image(image=image)
    blurred_edge_image = gaussian(edge_image, sigma=5)
    edges = canny(
        image=blurred_edge_image,
        sigma=1,
        low_threshold=blurred_edge_image.min(),
        high_threshold=blurred_edge_image.min() + (blurred_edge_image.max() - blurred_edge_image.min()) / 2
    )

    lines = probabilistic_hough_line(image=edges, line_gap=10, line_length=round(min(edge_image.shape[:2]) * 0.95))
    line = Line(x0=lines[0][0][0], y0=lines[0][0][1], x1=lines[0][1][0], y1=lines[0][1][1])

    xx, yy = np.meshgrid(np.arange(edge_image.shape[1]), np.arange(edge_image.shape[0]))
    point_data = [(get_shortest_line_point_dist(line=line, point=IntPoint(x=obj[0], y=obj[1])), obj[2])
                  for obj in np.vstack((xx.ravel(), yy.ravel(), edge_image.ravel())).T]
    point_data = sorted(point_data, key=lambda p: p[0])
    x, y = zip(*point_data)

    # Smooth the raw edge
    window_length = int(np.floor(np.divide(len(y), 100)))
    if window_length % 2 == 0:
        window_length += 1
    y_smooth = savgol_filter(x=y, window_length=window_length, polyorder=4)

    # Make sure that the edge has a higher value for positive x-values
    if np.mean(y_smooth[0:int(len(y_smooth) / 100)]) > np.mean(y_smooth[-int(len(y_smooth) / 100):]):
        # y = np.array(list(reversed(y)))
        y_smooth = np.array(list(reversed(y_smooth)))
        x = np.array([-el for el in reversed(x)])

    selection = (np.where((x > -20) & (x < 20)))
    y_smooth = y_smooth[selection]
    # y = y[selection]
    x = x[selection]

    dy = np.diff(y_smooth)
    # dx = np.diff(x)

    mean, std = norm.fit(dy)
    y_fit = norm.pdf(x, mean, std)

    mtf = abs(np.fft.rfft(y_fit))
    mtf = mtf / max(mtf)

    return MtfCurve(frequency=[i for i in range(len(mtf))], mtf=mtf)



from pathlib import Path
from setuptools import setup, find_packages

README = (Path(__file__).parent / 'README.md').read_text()

setup(
    name='dicom_image_tools',
    version='20.1.1',
    description='Python package for managing DICOM images from different modalities',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://dev.azure.com/bwkodex/DicomImageTools',
    author='Josef Lundman',
    author_email="josef@lundman.eu",
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'pydicom>=1.3.0',
        'numpy>=1.18.1',
        'scikit-image>=0.16.0',
        'scipy>=1.4.1'
    ],
    zip_safe=False
)
